"use strict";
var plan_1 = require("./plan");
var context_1 = require("./context");
var request_1 = require("./request");
var target_1 = require("./target");
var ERROR_MSGS = require("../constants/error_msgs");
var binding_type_1 = require("../bindings/binding_type");
var binding_count_1 = require("../bindings/binding_count");
var reflection_utils_1 = require("./reflection_utils");
var metadata_1 = require("../planning/metadata");
var METADATA_KEY = require("../constants/metadata_keys");
var serialization_1 = require("../utils/serialization");
function _createTarget(isMultiInject, targetType, serviceIdentifier, name, key, value) {
    var metadataKey = isMultiInject ? METADATA_KEY.MULTI_INJECT_TAG : METADATA_KEY.INJECT_TAG;
    var injectMetadata = new metadata_1.default(metadataKey, serviceIdentifier);
    var target = new target_1.default(targetType, name, serviceIdentifier, injectMetadata);
    if (key !== undefined) {
        var tagMetadata = new metadata_1.default(key, value);
        target.metadata.push(tagMetadata);
    }
    return target;
}
function _getActiveBindings(context, parentRequest, target) {
    var bindings = getBindings(context.container, target.serviceIdentifier);
    var activeBindings = [];
    if (bindings.length > 1) {
        activeBindings = bindings.filter(function (binding) {
            var request = new request_1.default(binding.serviceIdentifier, context, parentRequest, binding, target);
            return binding.constraint(request);
        });
    }
    else {
        activeBindings = bindings;
    }
    _validateActiveBindingCount(target.serviceIdentifier, activeBindings, target, context.container);
    return activeBindings;
}
function _validateActiveBindingCount(serviceIdentifier, bindings, target, container) {
    switch (bindings.length) {
        case binding_count_1.default.NoBindingsAvailable:
            var serviceIdentifierString = serialization_1.getServiceIdentifierAsString(serviceIdentifier);
            var msg = ERROR_MSGS.NOT_REGISTERED;
            msg += serialization_1.listMetadataForTarget(serviceIdentifierString, target);
            msg += serialization_1.listRegisteredBindingsForServiceIdentifier(container, serviceIdentifierString, getBindings);
            throw new Error(msg);
        case binding_count_1.default.OnlyOneBindingAvailable:
            if (target.isArray() === false) {
                return bindings;
            }
        case binding_count_1.default.MultipleBindingsAvailable:
        default:
            if (target.isArray() === false) {
                var serviceIdentifierString_1 = serialization_1.getServiceIdentifierAsString(serviceIdentifier), msg_1 = ERROR_MSGS.AMBIGUOUS_MATCH + " " + serviceIdentifierString_1;
                msg_1 += serialization_1.listRegisteredBindingsForServiceIdentifier(container, serviceIdentifierString_1, getBindings);
                throw new Error(msg_1);
            }
            else {
                return bindings;
            }
    }
}
function _createSubRequests(serviceIdentifier, context, parentRequest, target) {
    try {
        var activeBindings = void 0;
        var childRequest_1;
        if (parentRequest === null) {
            activeBindings = _getActiveBindings(context, null, target);
            childRequest_1 = new request_1.default(serviceIdentifier, context, null, activeBindings, target);
            var plan_2 = new plan_1.default(context, childRequest_1);
            context.addPlan(plan_2);
        }
        else {
            activeBindings = _getActiveBindings(context, parentRequest, target);
            childRequest_1 = parentRequest.addChildRequest(target.serviceIdentifier, activeBindings, target);
        }
        activeBindings.forEach(function (binding) {
            var subChildRequest = null;
            if (target.isArray()) {
                subChildRequest = childRequest_1.addChildRequest(binding.serviceIdentifier, binding, target);
            }
            else {
                subChildRequest = childRequest_1;
            }
            if (binding.type === binding_type_1.default.Instance) {
                var dependencies = reflection_utils_1.default(binding.implementationType);
                dependencies.forEach(function (dependency) {
                    _createSubRequests(dependency.serviceIdentifier, context, subChildRequest, dependency);
                });
            }
        });
    }
    catch (error) {
        if (error instanceof RangeError) {
            serialization_1.circularDependencyToException(parentRequest.parentContext.plan.rootRequest);
        }
        else {
            throw new Error(error.message);
        }
    }
}
function getBindings(container, serviceIdentifier) {
    var bindings = [];
    var bindingDictionary = container._bindingDictionary;
    if (bindingDictionary.hasKey(serviceIdentifier)) {
        bindings = bindingDictionary.get(serviceIdentifier);
    }
    else if (container.parent !== null) {
        bindings = getBindings(container.parent, serviceIdentifier);
    }
    return bindings;
}
function plan(container, isMultiInject, targetType, serviceIdentifier, key, value) {
    var context = new context_1.default(container);
    var target = _createTarget(isMultiInject, targetType, serviceIdentifier, "", key, value);
    _createSubRequests(serviceIdentifier, context, null, target);
    return context;
}
Object.defineProperty(exports, "__esModule", { value: true });
exports.default = plan;
